from __future__ import print_function
from Bio import Entrez
from lxml import etree
from os import path
from progressbar import (
    Bar,
    Counter,
    ETA,
    FileTransferSpeed,
    Percentage,
    ProgressBar,
)
from sys import stderr
from six.moves import urllib


# SRA XML xpath
XPATHS = {
    'name':
        '//EXPERIMENT_PACKAGE_SET/EXPERIMENT_PACKAGE/SAMPLE/@alias',
    'accession':
        '//EXPERIMENT_PACKAGE_SET/EXPERIMENT_PACKAGE/RUN_SET/RUN/@accession'
}


def human_readable_size(size, suffix='B'):
    '''Returns a human-readable representation of the size of something'''
    for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(size) < 1000.0:
            return "{sz:3.3f}{unt}{suf}".format(sz=size, unt=unit, suf=suffix)
        size /= 1000.0
    return "{sz:3.3f}{unt}{suf}".format(sz=size, unt='Y', suf=suffix)


def urlretrieve(url, filename, stream=stderr, retries=3):
    '''Downloads ``url`` to path ``filename``, silently if 'stream' is None'''

    class HumanReadableCounter(Counter):
        '''Counter() which prints a human readable size'''
        def update(self, pbar):
            return human_readable_size(pbar.currval)

    if stream is not None:
        print('Downloading ', filename, file=stream, end=', ')

    # Establish connection to server
    exc = None
    urlhandle = None
    while retries > 0 and urlhandle is None:
        try:
            urlhandle = urllib.request.urlopen(url, timeout=10)
        except urllib.error.URLError as exc:
            retries -= 1
            urlhandle = None
            exc = exc
    if urlhandle is None:
        if exc is not None:
            raise exc
        else:
            raise RuntimeError("Download of", url, "failed")

    # Get metadata, check if we need to download
    meta = urlhandle.info()
    file_size = int(meta.getheaders("Content-Length")[0])
    downloaded_size = 0
    # Check file size
    if path.exists(filename):
        downloaded_size = path.getsize(filename)
    if downloaded_size == file_size:
        if stream is not None:
            print('already downloaded!', file=stream)
        return

    # Set up the progress bar etc.
    if stream is not None:
        # On previous line
        print(human_readable_size(file_size), file=stream)
        # Widgets for the ProgressBar below
        widgets = [
            HumanReadableCounter(), ' ',
            FileTransferSpeed(), ' ',
            Bar(left='[', right=']'), ' ',
            Percentage(), ' ',
            ETA()
        ]
        pbar = ProgressBar(widgets=widgets, maxval=file_size)
        pbar.start()

    # Download the file
    file_size_dl = 0
    block_sz = 262144  # 256K
    with open(filename, 'wb') as fh:
        while True:
            buffer = urlhandle.read(block_sz)
            if not buffer:
                break
            fh.write(buffer)
            file_size_dl += len(buffer)
            if stream is not None:
                pbar.update(file_size_dl)
    if stream is not None:
        pbar.finish()


def accession_to_id(accession, force=False):
    '''Converts ``accession`` to being an id'''
    try:
        # Force the Entrez lookup by raising ValueError
        if force:
            raise ValueError
        # Try converting directly to an int, falling back to getting ID from
        # Entrez
        return int(accession)
    except ValueError:
        # Lookup numeric ID from accession
        term = '{}[Accession]'.format(accession)
        ids = esearch_ids(db='sra', term=term)
        if len(ids) == 1:
            return ids[0]
        else:
            raise ValueError("Couldn't get SRA UID of accession", accession)


def download_run(sra_id, outdir='.', stream=stderr, fmt='{acc}~{name}.sra'):
    '''Downloads run with SRA run uid ``sra_id`` to ``outdir``'''
    if stream is not None:
        print('Retrieving run info for run', sra_id, end='... ', file=stream)
    run_info = parse_run(sra_id)
    if stream is not None:
        print('done.', file=stream)
    url_template = ('ftp://ftp-trace.ncbi.nlm.nih.gov/sra/sra-instant/'
                    'reads/ByRun/sra/{leading3}/{leading6}/{all}/{all}.sra')
    run = run_info['accession']
    run_url = url_template.format(leading3=run[:3], leading6=run[:6], all=run)
    outfile = fmt.format(acc=run, name=run_info['name'], id=sra_id)
    outpath = path.join(outdir, outfile)
    urlretrieve(run_url, outpath)


def get_sample_runs(proj_id, stream=stderr):
    '''Gets a list of SRA run IDs from a given BioProject ID'''
    if stream is not None:
        print('Finding SRA runs for project', proj_id, end='... ', file=stream)
    term = '{}[BioProject]'.format(proj_id)
    sra_id_list = esearch_ids(db='sra', term=term)
    if stream is not None:
        print('found', len(sra_id_list), file=stream)
    return sra_id_list


def parse_run(sra_id):
    """
    Given an SRA id, gets the human-readable name and SRA Accession as a dict:

    {'id': sra_id, 'name': "sample_name", 'accession': 'SRR12345678'}

    sra_id should be a valid sra ID, as an int
    """
    sra_id = str(sra_id)
    handle = Entrez.efetch(db='sra', id=sra_id)
    xml_tree = etree.parse(handle)
    name = xml_tree.xpath(XPATHS['name'])[0]
    accession = xml_tree.xpath(XPATHS['accession'])[0]
    return {'id': sra_id, 'name': name, 'accession': accession}


def esearch_ids(**kwargs):
    '''
    Gets a list of IDs from an esearch query, handling mulitple requests.
    kwargs are passed directly to Bio.Entrez.esearch(), so consult their
    documentation, but 'db' and 'term' are required.
    '''
    handle = Entrez.esearch(**kwargs)
    rec = Entrez.read(handle)
    count = int(rec['Count'])
    id_list = rec['IdList']
    done = len(id_list)  # Num done so far
    while done < count:
        handle = Entrez.esearch(retstart=done, **kwargs)
        rec = Entrez.read(handle)
        id_list.extend(rec['IdList'])
        done += len(rec['IdList'])
    # Convert str-encoded ints to actual ints.
    id_list = map(int, id_list)
    return id_list


# versioneer
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



#!/usr/bin/env python
from __future__ import print_function
from Bio import Entrez
from docopt import docopt
from os import path
import sys

import srapy
from srapy import (
    get_sample_runs,
    download_run,
)


CLI_USAGE = """
USAGE:
    get-project-sras.py [-e EMAIL -d OUTDIR -F FMT] -p PROJECT_ID

OPTIONS:
    -e EMAIL        Your email, to provide to Bio.Entrez
                    [default: ] # defaults to empty string
    -d OUTDIR       Output directory, must exist. [default: .]
    -F FMT          Filename format. Fields 'name', 'id', and 'acc' are
                    recognised. Use python string formatting syntax.
                    [default: {acc}~{name}.sra]
    -p PROJECT_ID   BioProject ID
"""


def main(argv=sys.argv[1:]):
    opts = docopt(CLI_USAGE, argv=argv)
    proj_id = int(opts['-p'])
    Entrez.email = opts['-e']
    outdir = opts['-d']
    fname_fmt = opts['-F']

    if not path.isdir(outdir):
        print("ERROR: output directory '{}' doesn't exitst".format(outdir),
              file=sys.stderr)
        exit(1)

    runs = get_sample_runs(proj_id)
    # add a line between getting run list & downloading
    print(file=sys.stderr)
    for idx, sra_id in enumerate(runs):
        print("Downloading run", idx + 1, "of", len(runs), file=sys.stderr)
        download_run(sra_id, outdir=outdir, fmt=fname_fmt)
        print(file=sys.stderr)  # Extra newline


if __name__ == "__main__":
    print("SRApy version", srapy.__version__, file=sys.stderr, end='\n\n')
    main()



from os import path
from shutil import rmtree
from tempfile import mkdtemp

import srapy
from nose.tools import (
    assert_raises,
)

OUTDIR = None


def setup():
    '''Setup function for this module'''
    srapy.Entrez.email = 'kevin.murray@anu.edu.au'

    global OUTDIR
    OUTDIR = mkdtemp('srapy-test')


def teardown():
    '''Tear-down function for this module'''
    global OUTDIR
    rmtree(OUTDIR)
    OUTDIR = None


def test_accession_to_id():
    '''Test srapy.acccession_to_id'''

    # accession
    sra_id = srapy.accession_to_id('SRR1999343')
    assert sra_id == 1466194, sra_id

    # str(id)
    sra_id = srapy.accession_to_id('1466194')
    assert sra_id == 1466194, sra_id

    # int(id)
    sra_id = srapy.accession_to_id(1466194)
    assert sra_id == 1466194, sra_id

    # bad accession
    with assert_raises(ValueError):
        sra_id = srapy.accession_to_id('notacc')


def test_esearch_ids():
    '''Test srapy.esearch_ids'''

    # Search with fewer IDs than 'retmax'
    id_list = sorted(srapy.esearch_ids(db='bioproject', term='1001 genomes'))
    assert id_list == [30811, 197428, 273563], id_list

    # Search with more IDs than 'retmax'
    sra_id_list = [
        1508680, 1508681, 1508682, 1508683, 1508685, 1508686, 1508687, 1508688,
        1508689, 1508690, 1508691, 1508692, 1508693, 1508694, 1508695, 1508696,
    ]
    id_list = sorted(srapy.esearch_ids(db='sra', term='PRJDB3021'))
    assert id_list == sra_id_list, id_list

    # Search with no results
    id_list = srapy.esearch_ids(db='sra', term='abasdkjfslakdjflskdjflasjdf')
    assert id_list == [], id_list


def test_urlretrieve():
    '''Test srapy.urlretrieve'''

    # Test downloading a file silently
    outfile = path.join(OUTDIR, 'test_1mb.dat')
    srapy.urlretrieve('http://ftp.iinet.net.au/1mb.dat', outfile, stream=None)
    assert path.exists(outfile), 'Download failed'



#!/usr/bin/env python
from setuptools import setup
from setuptools.command.test import test as TestCommand
import versioneer


# Inspired by the example at https://pytest.org/latest/goodpractises.html
class NoseCommand(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # Run nose ensuring that argv simulates running nosetests directly
        import nose
        nose.run_exit(argv=['nosetests'])

versioneer.VCS = 'git'
versioneer.versionfile_source = 'srapy/_version.py'
versioneer.versionfile_build = 'srapy/_version.py'
versioneer.tag_prefix = ''
versioneer.parentdir_prefix = 'srapy-'

desc = """
SRApy: Scripts to download SRA files
"""

setup_requires = [
    'nose>=1.3,<1.4',
    'coverage>=3.7,<3.8',
]

install_requires = [
    'six>=1.9,<2.0',
    'biopython==1.65',
    'docopt>=0.6,<0.7',
    'lxml>=3.4,<3.5',
    'progressbar2>=2.7,<2.8',
]

test_requires = [
    'pep8>=1.6,<1.7',
    'pylint>=1.4,<1.5',
]

command_classes=versioneer.get_cmdclass()
command_classes['test'] =  NoseCommand

setup(
    name="srapy",
    packages=['srapy', ],
    scripts=[
        'scripts/get-project-sras.py',
        'scripts/get-run.py',
    ],
    version=versioneer.get_version(),
    cmdclass=command_classes,
    install_requires=install_requires,
    tests_require=test_requires,
    setup_requires=setup_requires,
    description=desc,
    author="Kevin Murray",
    author_email="spam@kdmurray.id.au",
    url="https://github.com/kdmurray91/srapy",
    keywords=[
        "bioinformatics",
        "SRA",
        "Next-gen Sequencing",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: GNU General Public License v3 or later " +
            "(GPLv3+)",
    ],
    test_suite="test",
)



#!/usr/bin/env python
from __future__ import print_function
from Bio import Entrez
from docopt import docopt
from os import path
import sys

import srapy
from srapy import (
    accession_to_id,
    download_run,
)


CLI_USAGE = """
USAGE:
    get-run.py [-e EMAIL -d OUTDIR -F FMT -a] (-i SRA_ID | -f FILE)

OPTIONS:
    -e EMAIL        Your email, to provide to Bio.Entrez
                    [default: ] # defaults to empty string
    -d OUTDIR       Output directory, must exist. [default: .]
    -a              The IDs are accessions. Unless '-a' is specified, the
                    identifier is assumed to be an SRA ID if it is numeric,
                    otherwise is interpreted as an accession. This option
                    forces the ID to be interpreted as an accession.
                    [default: False]
    -F FMT          Filename format. Fields 'name', 'id', and 'acc' are
                    recognised. Use python string formatting syntax.
                    [default: {acc}~{name}.sra]
    -i SRA_ID       A single identifier to download. See above for
                    interpretation
    -f FILE         New-line delimited list of identifiers
"""


def main(argv=sys.argv[1:]):
    opts = docopt(CLI_USAGE, argv=argv)
    Entrez.email = opts['-e']
    outdir = opts['-d']
    fname_fmt = opts['-F']

    if not path.isdir(outdir):
        print("ERROR: output directory '{}' doesn't exitst".format(outdir),
              file=sys.stderr)
        exit(1)

    ids = []

    if opts['-i']:
        ids.append(opts['-i'])
    else:
        with open(opts['-f']) as fh:
            for line in fh:
                ids.append(line.strip())

    uids = []
    bad_ids = []
    print("Getting accession UIDs ...", end='\r', file=sys.stderr)
    for idx, id_or_acc in enumerate(ids):
        try:
            uids.append(accession_to_id(id_or_acc, force=opts['-a']))
            print("Getting accession UIDs ...", idx + 1, "of", len(ids),
                  end='\r', file=sys.stderr)
        except ValueError:
            print("\nWARNING: Skipping non-existant accession", id_or_acc,
                  end='\n', file=sys.stderr)
            bad_ids.append(id_or_acc)
    print("\33[2KGetting accession UIDs ... Done", end='\n\n', file=sys.stderr)

    for idx, sra_id in enumerate(uids):
        print("Downloading run", idx + 1, "of", len(uids), file=sys.stderr)
        download_run(sra_id, outdir=outdir, fmt=fname_fmt)
        print(file=sys.stderr)  # Extra newline between records

    print("Finished!", file=sys.stderr)
    if len(bad_ids) > 0:
        print("The following accessions or IDs were skipped:", file=sys.stderr)
        for id in bad_ids:
            print("\t", id, file=sys.stderr)


if __name__ == "__main__":
    print("SRApy version", srapy.__version__, file=sys.stderr, end='\n\n')
    main()
